"""
Provides functionality for accessing resources on Posix file systems.
"""
import posixpath

from stor import base
from stor import utils


class PosixPath(base.FileSystemPath):
    """Represents a posix path.

    This class inherits a vendored path.py ``Path`` object and
    overrides methods for cross compatibility with swift.

    This class overrides ``Path`` object's
    ``module`` attribute and sets it to ``posixpath`` to ensure
    that it always represents a posix path.
    """
    path_module = posixpath

    def list(self):
        """
        List all files and directories under a path.

        Returns:
            List[str]: A list of all files and directories.
        """
        return utils.walk_files_and_dirs([self]).keys()

    def walkfiles(self, pattern=None):
        """Iterate over files recursively.

        Args:
            pattern (str, optional): Limits the results to files
                with names that match the pattern.  For example,
                ``mydir.walkfiles('*.tmp')`` yields only files with the ``.tmp``
                extension.

        Returns:
            Iter[Path]: Files recursively under the path
        """
        for f in self.list():
            if pattern is None or f.fnmatch(pattern):
                yield f



import functools
import logging
import time

logger = logging.getLogger(__name__)

EXCEPTIONS_ARG = "_exceptions"
INITIAL_SLEEP_ARG = "_initial_sleep"
TOTAL_RETRIES_ARG = "_total_retries"
SLEEP_FUNCTION_ARG = "_sleep_function"
CLEANUP_FUNCTION_ARG = "_cleanup_function"
IS_RETRY_OK_FUNCTION_ARG = "_is_retry_ok_function"

DEFAULT_EXCEPTIONS = (Exception,)
DEFAULT_INITIAL_SLEEP = 1
DEFAULT_TOTAL_RETRIES = 5
DEFAULT_SLEEP_FUNCTION = lambda t, attempt: t * 2
DEFAULT_CLEANUP_FUNCTION = None
DEFAULT_IS_RETRY_OK_FUNCTION = lambda error: True


def with_backoff(
        func=None,
        exceptions=DEFAULT_EXCEPTIONS,
        initial_sleep=DEFAULT_INITIAL_SLEEP,
        retries=DEFAULT_TOTAL_RETRIES,
        sleep_function=DEFAULT_SLEEP_FUNCTION,
        cleanup_function=DEFAULT_CLEANUP_FUNCTION,
        is_retry_ok_function=DEFAULT_IS_RETRY_OK_FUNCTION):
    """
    Decorator that retries a function with exponential backoff.

    Args:

        func (function): The function to decorate. Usually this is not needed;
            with_backoff is generally used as a decorator.
        exceptions (Exception or tuple of Exceptions): The exception or
            exceptions that will trigger a retry. Any other exceptions will
            raise immediately.
        initial_sleep (int): The initial sleep time after a failure.
        retries (int): The total number of times to retry.
        sleep_function (function(int, int) -> int): A function that increases
            sleep time each run. This function needs to take two integer
            arguments (time slept last attempt, attempt number) and return an
            integer. By default we simply multiply "t" by two every time.
        cleanup_function (function()): This is a function that takes no
            arguments. If provided, it will be run after each failure. It is
            intended to clean up after a failed call to the decorated
            function (temp files on disk, database connections, etc.)
        is_retry_ok_function (function(Exception) -> bool): This function takes
            a single argument, the exception raised during the latest retry of
            the decorated function. This function evaluates the raised
            exceptions and determines if the retry process should be
            terminated immediately or should continue normally. If this
            function returns True, the retry process will continue. If False is
            returned, the retry process will end immediately and the exception
            will be raised.

    Returns:

        A decorator that can be used to decorate unreliable functions (if
        called with no "func" argument

        A decorated function (if used as a decorator)

    All keyword arguments are optional. Set `exceptions` to limit which
    exceptions retry, by default it will retry on all exceptions.

    If you have an unreliable Python function or method, it might be useful to
    decorate it thusly:

    >>>@with_backoff
    >>>def some_operation_that_might_fail():
    >>>    import random
    >>>    assert random.randint(0, 10) > 5

    >>>@with_backoff(exceptions=AssertionError, retries=10)
    >>>def some_otheroperation_that_might_fail():
    >>>    import random
    >>>    assert random.randint(0, 10) > 5
    """

    wrapper_kwargs = {
        EXCEPTIONS_ARG: exceptions,
        INITIAL_SLEEP_ARG: initial_sleep,
        TOTAL_RETRIES_ARG: retries,
        SLEEP_FUNCTION_ARG: sleep_function,
        CLEANUP_FUNCTION_ARG: cleanup_function,
        IS_RETRY_OK_FUNCTION_ARG: is_retry_ok_function,
    }

    def decorated(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            kwargs.update(dict(wrapper_kwargs))
            return _backoff(f, *args, **kwargs)

        return inner

    if callable(func):
        return decorated(func)

    if func is not None:
        raise TypeError(
            "First positional argument, if provided, must be callable.")

    return decorated


def _backoff(f, *args, **kwargs):
    exceptions = kwargs.pop(EXCEPTIONS_ARG, DEFAULT_EXCEPTIONS)
    initial_sleep = kwargs.pop(INITIAL_SLEEP_ARG, DEFAULT_INITIAL_SLEEP)
    total_retries = kwargs.pop(TOTAL_RETRIES_ARG)
    sleep_function = kwargs.pop(SLEEP_FUNCTION_ARG, DEFAULT_SLEEP_FUNCTION)
    cleanup_function = kwargs.pop(
        CLEANUP_FUNCTION_ARG,
        DEFAULT_CLEANUP_FUNCTION)
    is_retry_ok_function = kwargs.pop(
        IS_RETRY_OK_FUNCTION_ARG,
        DEFAULT_IS_RETRY_OK_FUNCTION)

    sleep_time = initial_sleep
    for retry in range(total_retries):
        try:
            return f(*args, **kwargs)
        except exceptions as error:
            if not is_retry_ok_function(error):
                raise error
            time.sleep(sleep_time)
            sleep_time = sleep_function(sleep_time, retry)
            if cleanup_function is not None:
                cleanup_function()

    return f(*args, **kwargs)


import ast
import copy
from ConfigParser import SafeConfigParser
import os
import threading

CONFIG_FILE = 'default.cfg'
USER_CONFIG_FILE = '~/.stor.cfg'

_ENV_VARS = {
    'swift': {
        'username': 'OS_USERNAME',
        'password': 'OS_PASSWORD',
        'auth_url': 'OS_AUTH_URL',
        'temp_url_key': 'OS_TEMP_URL_KEY',
        'num_retries': 'OS_NUM_RETRIES'
    }
}
"""
A dictionary of options and their corresponding environment variables.

The top-level dictionary is a set of key-value pairs where keys correspond
to config sections and values are dictionaries where the keys are the option
names and values are the environment variables.s
"""

_global_settings = {}
thread_local = threading.local()


def _parse_config_val(value):
    try:
        return ast.literal_eval(value)
    except (SyntaxError, ValueError):
        return value


def _get_env_vars():
    """
    Update settings with environment variables, if applicable.

    Currently handles swift credentials.
    """
    new_settings = {}
    for section in _ENV_VARS:
        options = {}
        for option in _ENV_VARS[section]:
            if _ENV_VARS[section][option] in os.environ:
                options[option] = _parse_config_val(os.environ.get(_ENV_VARS[section][option]))
        new_settings[section] = options
    update(new_settings)


def parse_config_file(filename):
    """
    Parses a configuration file and returns a settings dictionary.

    Args:
        filename (str): File to read configuration settings from.

    Returns:
        dict: A dictionary of settings options.
    """
    parser = SafeConfigParser()

    with open(filename) as fp:
        parser.readfp(fp)

    settings = {
        section: {
            item[0]: _parse_config_val(item[1])
            for item in parser.items(section)
        }
        for section in parser.sections()
    }

    return settings


def _initialize():
    """
    Initialize global settings from configuration file. The configuration file
    **must** define all required settings, otherwise `stor` will not work.
    Every time this function is called it overwrites existing ``_global_settings``.
    When trying to update or change ``_global_settings``, `update` or `use` should
    be called instead.

    Also looks in the user's home directory for a custom configuration
    specified in ``~/.stor.cfg``.

    Defaults to reading from the default configuration file ``default.cfg``.

    Args:
        filename (str): File to read initial default configuration settings from.

    Returns:
        None
    """
    _global_settings.clear()
    default_cfg = os.path.join(os.path.dirname(__file__), CONFIG_FILE)
    update(parse_config_file(default_cfg), validate=False)
    custom_cfg = os.path.expanduser(USER_CONFIG_FILE)
    if os.path.exists(custom_cfg):
        update(parse_config_file(custom_cfg))
    _get_env_vars()


def _update(d, updates, validate=True):
    """
    Updates a nested dictionary with given dictionary

    If validate is set to True, the key being updated must already exist
    in the dictionary.
    """
    for key, value in updates.iteritems():
        if type(value) is dict:
            if key not in d or not type(d[key]) is dict:
                if validate:
                    raise ValueError('\'%s\' is not a valid setting' % key)
                d[key] = {}
            _update(d[key], value, validate)
        else:
            if validate and key not in d:
                raise ValueError('\'%s\' is not a valid setting' % key)
            d[key] = value


def get():
    """
    Returns a deep copy of global settings as a dictionary.

    This function should always be used rather than accessing
    ``global_settings`` directly.
    """
    try:
        return copy.deepcopy(thread_local.settings)
    except AttributeError:
        return copy.deepcopy(_global_settings)


def update(settings=None,
           # not documented
           validate=True):
    """
    Updates global settings permanently (in place).

    Arguments:
        settings (dict): A nested dictionary of settings options.

    Returns:
        None
    """
    if hasattr(thread_local, 'settings'):
        raise RuntimeError('update() cannot be called from within a settings context manager')
    if settings:
        _update(_global_settings, settings, validate=validate)


class _Use(object):
    """
    Context manager for temporarily modifying settings.
    """
    def __init__(self, settings=None):
        settings = settings or {}

        if hasattr(thread_local, 'settings'):
            self.old_settings = get()
        else:
            self.old_settings = None
        self.temp_settings = get()
        _update(self.temp_settings, settings)
        thread_local.settings = self.temp_settings

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        if self.old_settings:
            thread_local.settings = self.old_settings
        else:
            del thread_local.settings

#: Context manager for temporarily modifying settings.
#:
#: Arguments:
#:   settings (dict): A nested dictionary of settings options.
#:
#: Example:
#:      >>> from stor import settings
#:      >>> with settings.use({'swift:upload': {'object_threads': 20}}):
#:      >>>     # do something here
use = _Use



"""
Provides exceptions classes thrown by various utilities of stor
"""


class RemoteError(Exception):
    """
    The top-level exception thrown for any errors from remote services
    (eg. swift, S3).

    The 'caught_exception' attribute of the exception must be examined in order
    to inspect the exception thrown by the swift or S3 service. A swift exception
    can either be a ``SwiftError`` (thrown by ``swiftclient.service``) or a
    ``ClientError`` (thrown by ``swiftclient.client``). A S3 exception can be
    a ``ClientError`` (thrown by ``botocore.client``).
    """
    def __init__(self, message, caught_exception=None):
        super(RemoteError, self).__init__(message)
        self.caught_exception = caught_exception


class NotFoundError(RemoteError):
    """Thrown when a 404 response is returned."""
    pass


class UnauthorizedError(RemoteError):
    """Thrown when a 403 response is returned."""
    pass


class UnavailableError(RemoteError):
    """Thrown when a 503 response is returned."""
    pass


class ConditionNotMetError(RemoteError):
    """Thrown when a condition is not met."""
    pass


class FailedTransferError(RemoteError):
    """Thrown when a file transfer fails."""
    pass


class FailedUploadError(FailedTransferError):
    """Thrown when an upload fails."""
    pass


class FailedDownloadError(FailedTransferError):
    """Thrown when a download fails."""
    pass



import mock
from stor import s3
from stor.s3 import S3Path
from stor.swift import SwiftPath
from stor import settings
import unittest


class SwiftTestMixin(object):
    """A mixin with helpers for mocking out swift.

    SwiftTestMixin should be used to create base test classes for anything
    that accesses swift.
    """
    def disable_get_swift_service_mock(self):
        """Disables the mock for getting the swift service.
        """
        try:
            self._get_swift_patcher.stop()
        except RuntimeError:
            # If the user disables the mock, the mock will try
            # to be stopped on test cleanup. Disable errors from that
            pass

    def setup_swift_mocks(self):
        """Sets all of the relevant mocks for Swift communication.

        If you are testing outside of this library, you should either mock
        swift object methods or you should focus on manipulating return value
        of mock_swift.

        The following variables are set up when calling this:

        - mock_swift_service: A mock of the SwiftService class defined in
          swiftclient.service.

        - mock_swift_get_conn: A mock of the get_conn function in the
          swiftclient.service module

        - mock_swift_conn: A mock of the SwiftConnection returned by
          get_conn

        - mock_swift_get_auth_keystone: mock of the get_keystone_auth function
          that caches identity credentials

        - mock_get_swift_service: A mock of the _get_swift_service method of
          SwiftPath

        - mock_swift: A mock of the SwiftService instance returned by
          _get_swift_service in SwiftPath
        """
        # Ensure that SwiftService will never be instantiated in tests
        swift_service_patcher = mock.patch('swiftclient.service.SwiftService',
                                           autospec=True)
        self.addCleanup(swift_service_patcher.stop)
        self.mock_swift_service = swift_service_patcher.start()

        # Ensure that SwiftConnections will never be instantiated in tests
        swift_get_conn_patcher = mock.patch('swiftclient.service.get_conn',
                                            autospec=True)
        self.addCleanup(swift_get_conn_patcher.stop)
        self.mock_swift_get_conn = swift_get_conn_patcher.start()
        self.mock_swift_conn = mock.Mock()
        self.mock_swift_get_conn.return_value = self.mock_swift_conn

        # Ensure that no keystone auth calls will go out
        swift_keystone_mock_patcher = mock.patch('swiftclient.client.get_auth_keystone',
                                                 autospec=True)
        self.addCleanup(swift_keystone_mock_patcher.stop)
        self.mock_swift_get_auth_keystone = swift_keystone_mock_patcher.start()
        self.mock_swift_get_auth_keystone.return_value = ('dummy_storage_url', 'dummy_auth_token')

        # This is the mock that will always be returned by _get_swift_service.
        # The user can mock out any swift methods on this mock
        self.mock_swift = mock.Mock()
        self._get_swift_patcher = mock.patch.object(SwiftPath,
                                                    '_get_swift_service',
                                                    autospec=True)
        self.addCleanup(self.disable_get_swift_service_mock)
        self.mock_get_swift_service = self._get_swift_patcher.start()
        self.mock_get_swift_service.return_value = self.mock_swift

        # ensures we never cache data between tests
        _cache_patcher = mock.patch.dict('stor.swift._cached_auth_token_map', clear=True)
        self.addCleanup(_cache_patcher.stop)
        _cache_patcher.start()

    def assertSwiftListResultsEqual(self, r1, r2):
        """
        Swift list resolves duplicates, so the ordering of the results are not
        always the same as what the swift client returns. Compare results as
        sorted lists
        """
        self.assertEquals(sorted(r1), sorted(r2))


class S3TestMixin(object):
    """A mixin with helpers for mocking out S3.

    S3TestMixin should be used to create base test classes for anything
    that accesses S3.
    """
    def disable_get_s3_client_mock(self):
        """Disables the mock for getting the S3 client."""
        try:
            self._get_s3_client_patcher.stop()
        except RuntimeError:
            # If the user disables the mock, the mock will try
            # to be stopped on test cleanup. Disable errors from that
            pass

    def disable_get_s3_iterator_mock(self):
        """Disables the mock for getting the S3 iterator."""
        try:
            self._get_s3_iterator_patcher.stop()
        except RuntimeError:
            pass

    def setup_s3_mocks(self):
        """Sets all of the relevant mocks for S3 communication.

        If you are testing outside of this library, you should either mock
        S3 client methods or you should focus on manipulating return value
        of mock_s3.

        Tests of methods that directly make API calls via _s3_client_call should
        mock the return values of the API calls on mock_s3. Tests of methods that
        do not directly make the API calls should mock any S3Path methods being called.

        The following variables are set up when calling this:

        - mock_s3_client: A mock of the Client instance returned by boto3.client

        - mock_s3: A mock of the Client instance returned by _get_s3_client in S3Path.

        - mock_get_s3_client: A mock of the _get_s3_client method in S3Path.

        - mock_get_s3_iterator: A mock of the _get_s3_iterator method in S3Path.

        - mock_s3_iterator: A mock of the iterable object returned by _get_s3_iterator in S3Path.

        - mock_s3_transfer: A mock of the Transfer instance returned by S3Transfer

        - mock_get_s3_transfer: A mock of the boto3.s3.transfer.S3Transfer object
        """
        # Ensure that the S3 session will never be instantiated in tests
        s3_session_patcher = mock.patch('boto3.session.Session', autospec=True)
        self.addCleanup(s3_session_patcher.stop)
        self.mock_s3_session = s3_session_patcher.start()

        # This is the mock returned by _get_s3_client.
        # User can mock s3 methods on this mock.
        self.mock_s3 = mock.Mock()
        self._get_s3_client_patcher = mock.patch('stor.s3._get_s3_client',
                                                 autospec=True)
        self.addCleanup(self.disable_get_s3_client_mock)
        self.mock_get_s3_client = self._get_s3_client_patcher.start()
        self.mock_get_s3_client.return_value = self.mock_s3

        # This is the mock returned by _get_s3_iterator.
        # User should modify the __iter__.return_value property to specify return values.
        self.mock_s3_iterator = mock.MagicMock()
        self._get_s3_iterator_patcher = mock.patch.object(S3Path, '_get_s3_iterator',
                                                          autospec=True)
        self.addCleanup(self.disable_get_s3_iterator_mock)
        self.mock_get_s3_iterator = self._get_s3_iterator_patcher.start()
        self.mock_get_s3_iterator.return_value = self.mock_s3_iterator

        # Ensure that an S3Transfer object will never be instantiated in tests.
        # User can mock methods associated with S3Transfer on this mock.
        self.mock_s3_transfer = mock.Mock()
        s3_transfer_patcher = mock.patch('stor.s3.S3Transfer', autospec=True)
        self.addCleanup(s3_transfer_patcher.stop)
        self.mock_get_s3_transfer = s3_transfer_patcher.start()
        self.mock_get_s3_transfer.return_value = self.mock_s3_transfer

        # Mock the TransferConfig object
        s3_transfer_config_patcher = mock.patch('stor.s3.TransferConfig',
                                                autospec=True)
        self.addCleanup(s3_transfer_config_patcher.stop)
        self.mock_get_s3_transfer_config = s3_transfer_config_patcher.start()


class SwiftTestCase(unittest.TestCase, SwiftTestMixin):
    """A TestCase class that sets up swift mocks and provides additional assertions"""
    def setUp(self):
        super(SwiftTestCase, self).setUp()
        self.setup_swift_mocks()

        # make sure swift credentials aren't included
        settings.update({
            'swift': {
                'username': '__dummy__',
                'password': '__dummy__',
                'auth_url': '__dummy__'
            }
        })


class S3TestCase(unittest.TestCase, S3TestMixin):
    """A TestCase class that sets up S3 mocks"""
    def setUp(self):
        super(S3TestCase, self).setUp()
        self.setup_s3_mocks()
        try:
            del s3._thread_local.s3_transfer
            del s3._thread_local.s3_transfer_config
        except AttributeError:
            pass



"""
stor
====

Stor is a library that aims to make it easy to write code that works
with both local posix filesystems and Swift Object Based Storage. In general,
you should be able to replace most uses of ``os.path`` and ``open`` with::

    import stor as path
    from stor import open

And your code will work either with posix paths or swift paths (defined as
strings in the format ``swift://<TENANT>/<CONTAINER>/<OBJECT>``).  Stor
also provides an object-oriented API similar to Python 3's new pathlib,
accessible via ``stor.Path``.

stor is heavily inspired by / based on the path.py library, but
modified to avoid the need to know whether you have a Path or a string for most
functions.

See `stor.swift` for more information on Swift-specific functionality.
"""
import pkg_resources

from stor.utils import copy
from stor.utils import copytree
from stor.utils import is_filesystem_path
from stor.utils import is_swift_path
from stor.utils import NamedTemporaryDirectory
from stor.base import Path
from stor import settings


# TODO: Compile this - costs ~700us to do this on import
__version__ = pkg_resources.get_distribution('stor').version


settings._initialize()


def _delegate_to_path(name):
    def wrapper(path, *args, **kwargs):
        f = getattr(Path(path), name)
        return f(*args, **kwargs)
    wrapper.__doc__ = getattr(Path, name).__doc__
    wrapper.__name__ = name
    return wrapper

# extra compat!
open = _delegate_to_path('open')
abspath = _delegate_to_path('abspath')
normcase = _delegate_to_path('normcase')
normpath = _delegate_to_path('normpath')
realpath = _delegate_to_path('realpath')
expanduser = _delegate_to_path('expanduser')
expandvars = _delegate_to_path('expandvars')
dirname = _delegate_to_path('dirname')
basename = _delegate_to_path('basename')
expand = _delegate_to_path('expand')
# these two have special names to not collide with builtins
join = _delegate_to_path('joinpath')
split = _delegate_to_path('splitpath')
splitext = _delegate_to_path('splitext')
list = _delegate_to_path('list')
listdir = _delegate_to_path('listdir')
glob = _delegate_to_path('glob')
exists = _delegate_to_path('exists')
isabs = _delegate_to_path('isabs')
isdir = _delegate_to_path('isdir')
isfile = _delegate_to_path('isfile')
islink = _delegate_to_path('islink')
ismount = _delegate_to_path('ismount')
getsize = _delegate_to_path('getsize')
remove = _delegate_to_path('remove')
rmtree = _delegate_to_path('rmtree')
walkfiles = _delegate_to_path('walkfiles')


def listpath(pth):
    import warnings

    # DeprecationWarnings are hidden by default. We want to get rid of this
    # sooner rather than later.
    warnings.warn('Using the ``stor.listpath()`` function directly is'
                  ' deprecated, use ``stor.list()`` instead', UserWarning)
    return list(pth)


def path(pth):  # pragma: no cover
    import warnings

    # DeprecationWarnings are hidden by default. We want to get rid of this
    # sooner rather than later.
    warnings.warn('Using the ``path()`` function directly is deprecated -'
                  ' either use stor.Path or the functional API'
                  ' directly', UserWarning)
    return Path(pth)


__all__ = [
    'abspath',
    'normcase',
    'normpath',
    'realpath',
    'expanduser',
    'expandvars',
    'dirname',
    'basename',
    'expand',
    'join',
    'split',
    'splitext',
    'listdir',
    'glob',
    'exists',
    'isabs',
    'isdir',
    'isfile',
    'islink',
    'ismount',
    'getsize',
    'copy',
    'copytree',
    'remove',
    'rmtree',
    'walkfiles',
    'is_filesystem_path',
    'is_swift_path',
    'NamedTemporaryDirectory',
]



from contextlib import contextmanager
import datetime
import errno
import logging
import os
import shlex
import shutil
from subprocess import check_call
import tempfile
from stor import exceptions

logger = logging.getLogger(__name__)

# Name for the data manifest file when using the use_manifest option
# for upload/download
DATA_MANIFEST_FILE_NAME = '.data_manifest.csv'


def str_to_bytes(s):
    """
    Converts a given string into an integer representing bytes
    where G is gigabytes, M is megabytes, K is kilobytes, and B
    is bytes.
    """
    if type(s) is int:
        return s

    units = {'B': 1, 'K': 1024, 'M': 1024 ** 2, 'G': 1024 ** 3}
    if len(s) < 2:
        raise ValueError('invalid size')
    order = s[-1]
    try:
        return units[order] * int(s[:-1])
    except ValueError:
        raise ValueError('invalid size')
    except KeyError:
        raise ValueError('invalid units')


def file_name_to_object_name(p):
    """Given a file path, construct its object name.

    Any relative or absolute directory markers at the beginning of
    the path will be stripped, for example::

        ../../my_file -> my_file
        ./my_dir -> my_dir
        .hidden_dir/file -> .hidden_dir/file
        /absolute_dir -> absolute_dir

    Note that windows paths will have their back slashes changed to
    forward slashes::

        C:\\my\\windows\\file -> my/windows/file

    Args:
        p (str): The input path

    Returns:
        PosixPath: The object name. An empty path will be returned in
            the case of the input path only consisting of absolute
            or relative directory markers (i.e. '/' -> '', './' -> '')
    """
    from stor import Path
    from stor.posix import PosixPath

    p_parts = Path(p).expand().splitdrive()[1].split(os.path.sep)
    obj_start = next((i for i, part in enumerate(p_parts) if part not in ('', '..', '.')), None)
    return PosixPath.parts_class('/'.join(p_parts[obj_start:]) if obj_start is not None else '')


def make_dest_dir(dest):
    """Make directories if they do not already exist.

    Raises:
        OSError: An error occurred while creating the directories.
            A specific exception is if a directory that is being created already exists as a file.
    """
    dest = os.path.abspath(dest)
    try:
        os.makedirs(dest)
    except OSError as exc:
        if exc.errno == errno.ENOTDIR:
            raise OSError(errno.ENOTDIR,
                          'a parent directory of \'%s\' already exists as a file' % dest)
        elif exc.errno != errno.EEXIST or os.path.isfile(dest):
            raise


def with_trailing_slash(p):
    """Returns a path with a single trailing slash or None if not a path"""
    if not p:
        return p
    return type(p)(p.rstrip('/') + '/')


def has_trailing_slash(p):
    """Checks if a path has a single trailing slash"""
    if not p:
        return False
    return str(p)[-1] == '/'


def remove_trailing_slash(p):
    """Returns a path with trailing slashes removed or None if not a path"""
    if not p:
        return p
    return type(p)(p.rstrip('/'))


def validate_condition(condition):
    """Verifies condition is a function that takes one argument"""
    if condition is None:
        return
    if not (hasattr(condition, '__call__') and hasattr(condition, '__code__')):
        raise ValueError('condition must be callable')
    if condition.__code__.co_argcount != 1:
        raise ValueError('condition must take exactly one argument')


def check_condition(condition, results):
    """Checks the results against the condition.

    Raises:
        ConditionNotMetError: If the condition returns False
    """
    if condition is None:
        return

    condition_met = condition(results)
    if not condition_met:
        raise exceptions.ConditionNotMetError('condition not met')


def join_conditions(*conditions):
    def wrapper(results):
        return all(f(results) for f in conditions)
    return wrapper


def generate_and_save_data_manifest(manifest_dir, data_manifest_contents):
    """Generates a data manifest for a given directory and saves it.

    Args:
        manifest_dir (str): The directory in which the manifest will be saved
        data_manifest_contents (List[str]): The list of all objects that will
            be part of the manifest.
    """
    import stor
    from stor import Path

    manifest_file_name = Path(manifest_dir) / DATA_MANIFEST_FILE_NAME
    with stor.open(manifest_file_name, 'w') as out_file:
        contents = '\n'.join(data_manifest_contents) + '\n'
        out_file.write(contents)


def get_data_manifest_contents(manifest_dir):
    """Reads the manifest file and returns a set of expected files"""
    import stor

    manifest = manifest_dir / DATA_MANIFEST_FILE_NAME
    with stor.open(manifest, 'r') as manifest_file:
        return [
            f.strip() for f in manifest_file.readlines() if f.strip()
        ]


def validate_manifest_list(expected_objs, list_results):
    """
    Given a list of expected object names and results,
    verify all expected objects are in the listed results
    """
    listed_objs = {r.resource for r in list_results}
    return set(expected_objs).issubset(listed_objs)


def is_swift_path(p):
    """Determines if the path is a Swift path.

    All Swift paths start with swift://

    Args:
        p (str): The path string

    Returns:
        bool: True if p is a Swift path, False otherwise.
    """
    from stor.swift import SwiftPath
    return p.startswith(SwiftPath.drive)


def is_filesystem_path(p):
    """Determines if the path is posix or windows filesystem.

    Args:
        p (str): The path string

    Returns:
        bool: True if p is a Windows path, False otherwise.
    """
    return not (is_swift_path(p) or is_s3_path(p))


def is_s3_path(p):
    """Determines if the path is a S3 path.

    All S3 paths start with ``s3://``

    Args:
        p (str): The path string

    Returns
        bool: True if p is a S3 path, False otherwise.
    """
    from stor.s3 import S3Path
    return p.startswith(S3Path.drive)


def is_obs_path(p):
    """Determines if the path is an OBS path (an S3 or Swift path).

    Args:
        p (str): The path string

    Returns
        bool: True if p is an OBS path, False otherwise.
    """
    return is_s3_path(p) or is_swift_path(p)


def copy(source, dest, swift_retry_options=None):
    """Copies a source file to a destination file.

    Note that this utility can be called from either OBS, posix, or
    windows paths created with ``stor.Path``.

    Args:
        source (path|str): The source directory to copy from
        dest (path|str): The destination file or directory.
        swift_retry_options (dict): Optional retry arguments to use for swift
            upload or download. View the
            `swift module-level documentation <swiftretry>` for more
            information on retry arguments

    Examples:
        Copying a swift file to a local path behaves as follows::

            >>> import stor
            >>> swift_p = 'swift://tenant/container/dir/file.txt'
            >>> # file.txt will be copied to other_dir/other_file.txt
            >>> stor.copy(swift_p, 'other_dir/other_file.txt')

        Copying from a local path to swift behaves as follows::

            >>> from stor import Path
            >>> local_p = Path('my/local/file.txt')
            >>> # File will be uploaded to swift://tenant/container/dir/my_file.txt
            >>> local_p.copy('swift://tenant/container/dir/')

        Because of the ambiguity in whether a remote target is a file or directory, copy()
        will error on ambiguous paths.

            >>> local_p.copy('swift://tenant/container/dir')
            Traceback (most recent call last):
            ...
            ValueError: OBS destination must be file with extension or directory with slash
    """
    from stor import Path
    from stor.obs import OBSUploadObject

    source = Path(source)
    dest = Path(dest)
    swift_retry_options = swift_retry_options or {}
    if is_obs_path(source) and is_obs_path(dest):
        raise ValueError('cannot copy one OBS path to another OBS path')
    if is_obs_path(dest) and dest.is_ambiguous():
        raise ValueError('OBS destination must be file with extension or directory with slash')

    if is_filesystem_path(dest):
        if is_obs_path(source):
            dest_file = dest if not dest.isdir() else dest / source.name
            source.download_object(dest_file, **swift_retry_options)
        else:
            shutil.copy(source, dest)
    else:
        dest_file = dest if not dest.endswith('/') else dest / source.name
        if is_swift_path(dest) and not dest_file.parent.container:
            raise ValueError((
                'cannot copy to tenant "%s" and file '
                '"%s"' % (dest_file.parent, dest_file.name)
            ))
        dest_obj_name = Path(dest_file.parent.resource or '') / dest_file.name
        upload_obj = OBSUploadObject(source, dest_obj_name)
        dest_file.parent.upload([upload_obj],
                                **swift_retry_options)


def copytree(source, dest, copy_cmd=None, use_manifest=False, headers=None,
             condition=None, **retry_args):
    """Copies a source directory to a destination directory. Assumes that
    paths are capable of being copied to/from.

    Note that this function uses shutil.copytree by default, meaning
    that a posix or windows destination must not exist beforehand.

    For example, assume the following file hierarchy::

        a/
        - b/
        - - 1.txt

    Doing a copytree from ``a`` to a new posix destination of ``c`` is
    performed with::

        Path('a').copytree('c')

    The end result for c looks like::

        c/
        - b/
        - - 1.txt

    Note that the user can override which copy command is used for posix
    copies, and it is up to them to ensure that their code abides by the
    semantics of the provided copy command. This function has been tested
    in production using the default command of ``cp -r`` and using ``mcp -r``.

    Using OBS source and destinations work in a similar manner. Assume
    the destination is a swift path and we upload the same ``a`` folder::

        Path('a').copytree('swift://tenant/container/folder')

    The end swift result will have one object::

        Path('swift://tenant/container/folder/b/1.txt')

    Similarly one can do::

        Path('swift://tenant/container/folder/').copytree('c')

    The end result for c looks the same as the above posix example::

        c/
        - b/
        - - 1.txt

    Args:
        source (path|str): The source directory to copy from
        dest (path|str): The directory to copy to. Must not exist if
            its a posix directory
        copy_cmd (str): If copying to / from posix or windows, this command is
            used instead of shutil.copytree
        use_manifest (bool, default False): See `SwiftPath.upload` and
            `SwiftPath.download`.
        headers (List[str]): See `SwiftPath.upload`.

    Raises:
        ValueError: if two OBS paths are specified
        OSError: if destination is a posix path and it already exists
    """
    from stor import Path

    source = Path(source)
    dest = Path(dest)
    if is_obs_path(source) and is_obs_path(dest):
        raise ValueError('cannot copy one OBS path to another OBS path')
    from stor.windows import WindowsPath
    if is_obs_path(source) and isinstance(dest, WindowsPath):
        raise ValueError('OBS copytree to windows is not supported')

    if is_filesystem_path(dest):
        dest.expand().abspath().parent.makedirs_p()
        if is_obs_path(source):
            source.download(dest, use_manifest=use_manifest,
                            condition=condition, **retry_args)
        else:
            if copy_cmd:
                copy_cmd = shlex.split(copy_cmd)
                copy_cmd.extend([str(source.abspath().expand()),
                                 str(dest.abspath().expand())])
                logger.info('performing copy with command - %s', copy_cmd)
                check_call(copy_cmd)
            else:
                shutil.copytree(source, dest)
    else:
        with source:
            dest.upload(['.'], use_manifest=use_manifest, headers=headers,
                        condition=condition, **retry_args)


def walk_files_and_dirs(files_and_dirs):
    """Walk all files and directories.

    Args:
        files_and_dirs (List[str]): All file or directory names to walk.

    Returns:
        dict: All files and empty directories under files_and_dirs keyed to
            their size. Dictionaries have a size of 0

    Raises:
        ValueError: The provided upload name is not a file or a directory.

    Examples:
        >>> from stor.utils import walk_files_and_dirs
        >>> results = walk_files_and_dirs(['file_name', 'dir_name'])
        >>> print results
        ['file_name', 'dir_name/file1', 'dir_name/file2']
    """
    walked_upload_names_and_sizes = {}
    for name in files_and_dirs:
        if os.path.isfile(name):
            walked_upload_names_and_sizes[name] = os.path.getsize(name)
        elif os.path.isdir(name):
            for (root_dir, dir_names, file_names) in os.walk(name):
                if not (dir_names + file_names):
                    # Ensure that empty directories are uploaded as well
                    walked_upload_names_and_sizes[root_dir] = 0
                else:
                    for file_name in file_names:
                        full_name = os.path.join(root_dir, file_name)
                        walked_upload_names_and_sizes[full_name] = os.path.getsize(full_name)
        else:
            raise ValueError('file "%s" not found' % name)

    return walked_upload_names_and_sizes


@contextmanager
def NamedTemporaryDirectory(suffix='', prefix='tmp', dir=None,
                            change_dir=False):
    """Context manager for creating and deleting temporary directory.

    Mimics the behavior of tempfile.NamedTemporaryFile.

    Arguments:
        suffix (str): If specified, the dir name will end with it.
        prefix (str): If specified, the dir name will start with it,
            otherwise 'tmp' is used.
        dir (str): If specified, the dir will be created in this
            directory.
        change_dir (bool): If specified, will change to the temporary
            directory.

    Yields:
        Path: The temporary directory.

    Note:
        Name is CamelCase to match tempfile.NamedTemporaryFile.

    Examples:
        >>> from stor import NamedTemporaryDirectory
        >>> with NamedTemporaryDirectory() as d:
        >>>     # Do operations within "d", which will be deleted afterwards
    """
    from stor import Path

    tempdir = Path(tempfile.mkdtemp(suffix, prefix, dir))
    try:
        if change_dir:
            with tempdir:
                yield tempdir
        else:
            yield tempdir
    finally:
        tempdir.rmtree()


class ClassProperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class BaseProgressLogger(object):
    """Base class and methods for logging progress.

    Progress loggers that inherit this base class have the ability to
    track information related to the progress of a group of results.

    Users instantiate the logger as a context manager::

        for MyProgressLogger() as l:
            for r in results:
                l.add_result(r)

    When the progress logger is instantiated, the message returned
    from ``get_start_message`` is logged. As results are added, logs
    are printed at each result interval. For example, if the
    ``result_interval`` is 3, progress will be logged for every third result
    added.

    When exiting the context manager, the log message from ``get_finish_message``
    is returned.

    To accumulate more results, implement the ``update_progress`` method. For
    example, to track the amount of bytes tracked so far::

        def update_progress(self, result):
            self.total_bytes += result['bytes']

    Any custom results can be printed when implementing ``get_progress_message``.
    """
    def __init__(self, logger, level=logging.INFO, result_interval=10):
        self.logger = logger
        self.level = level
        self.result_interval = result_interval
        self.num_results = 0
        self.start_time = datetime.datetime.utcnow()

    def __enter__(self):
        start_msg = self.get_start_message()
        if start_msg:  # pragma: no cover
            self.logger.log(self.level, start_msg)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type is None:
            finish_msg = self.get_finish_message()
            if finish_msg:
                self.logger.log(self.level, finish_msg)

    def get_elapsed_time(self):
        return datetime.datetime.utcnow() - self.start_time

    def format_time(self, t):
        time_elapsed = datetime.datetime.utcnow() - self.start_time
        hours, remainder = divmod(time_elapsed.total_seconds(), 3600)
        minutes, seconds = divmod(remainder, 60)
        return '%d:%02d:%02d' % (hours, minutes, seconds)

    def get_start_message(self):
        return self.get_progress_message()

    def get_finish_message(self):
        return self.get_progress_message()

    def get_progress_message(self):
        raise NotImplementedError

    def update_progress(self, result):
        pass

    def add_result(self, result):
        """Adds a result to the progress logger and logs messages.

        Messages are logged every time the ``result_interval`` is met.
        """
        self.num_results += 1
        self.update_progress(result)
        if self.num_results % self.result_interval == 0:
            progress_msg = self.get_progress_message()
            if progress_msg:  # pragma: no cover
                self.logger.log(self.level, progress_msg)
